# -*- coding: utf-8 -*-
"""
This script is the cVAE for the publication 
Machine learning of optical properties of materials - predicting spectra from images and images from spectra
by
Helge S. Stein, Dan Guevara, Paul F. Newhouse, Edwin Soedarmadji, John M. Gregoire,
Joint Center for Artificial Photosynthesis
California Institute of Technology
Pasadena
California 91125 (USA)

correspondence should be adressed to :
stein@caltech.edu, gregoire@caltech.edu
"""


import numpy as np
from keras.layers import Input, Dense, Lambda, Reshape, Flatten, Dropout
from keras.layers import Conv1D, MaxPooling1D
from keras.layers.merge import concatenate
from keras.models import Model
from keras import backend as K
from keras.callbacks import EarlyStopping
from keras.utils import HDF5Matrix
import pickle
from keras import metrics
from keras.layers import BatchNormalization
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler, MinMaxScaler, QuantileTransformer


def load_data(path):
    #all this does now is to import data preprocessign is done later in the generator 
    X_train = HDF5Matrix(path, 'images_train')
    X_test = HDF5Matrix(path, 'images_test')

    y_train = HDF5Matrix(path, 'spectra_train')
    y_test = HDF5Matrix(path, 'spectra_test')


    return X_train, X_test, y_train, y_test

X_train, X_test, y_train, y_test = load_data(r'D:\dataset_publish\test_train_data.h5')

params = {
        'optimizer' : 'adam',
        'm' : 2048,
        'n_x' : X_train.shape[1:], #symmetric
        'n_y' : y_train.shape[1],
        'n_z' : 100,
        'nepoch' : 1000
        }

##  CONDITIONAL VARIATIONAL ENCODER

X = Input(shape=params['n_x'],name='inputimg')
cond = Input(shape=(params['n_y'], ),name='inputcond')

#encoder

cf = concatenate([Flatten()(X), cond])

mid = Dense(params['m'],activation='tanh')(BatchNormalization()(cf))

#here we use a non regularized linear activation to allow for zero meanunit variance
mu_a = Dense(params['n_z'],name='mu',activation='linear')(BatchNormalization()(mid))
log_sigma_a = Dense(params['n_z'],name='logsigma',activation='linear')(BatchNormalization()(mid))

def sample_z(args):
    mu, log_sigma = args
    z_rand = mu +  K.random_normal(K.shape(mu),mean=0.0, stddev=1.0) * K.exp(log_sigma/2)

    return K.in_train_phase(z_rand,mu)

# Sampling latent space via lambda layer
z = Lambda(sample_z, output_shape = (params['n_z'], ),
           name='zlayer')([mu_a, log_sigma_a])

# merge latent space with condition
z_cond = concatenate([z, cond],name='zcond') #took away BN here .. check . did because spectra looked strange

encoder = Model([X, cond], z_cond) #---------------------------------------------------------------------

decoder_input = Input(shape=(params['n_y']+params['n_z'], ),name='input_decoder')

decoder_dense_a = Dense(64*64*3,activation='sigmoid')(decoder_input)


outputs_a = Reshape(target_shape=(64,64,3,))(decoder_dense_a)

decoder = Model(decoder_input,outputs_a) #----------------------------------------------------------

encoder_use = Model([X, cond], [mu_a,log_sigma_a])

cvae = Model(inputs=encoder.input, outputs=decoder(encoder.output))

cvae.summary()

def vae_loss(x, out):
    xent_loss =  metrics.binary_crossentropy(x, out)*64*64*3
    kl_loss = - 0.5 * K.mean(1 + log_sigma_a - K.square(mu_a) - K.exp(log_sigma_a),axis=-1)/(10**6)
    #xent_loss and kl_loss are weighted via trail and error
    return (K.mean(xent_loss) + K.mean(kl_loss))

def recon_loss(x, out):
    return  metrics.binary_crossentropy(x, out)*64*64*3

def KL_loss(inp, outp):
    x_mean, x_log_var = mu_a,log_sigma_a
    kl_loss = - 0.5 * K.mean(1 + x_log_var - K.square(x_mean) - K.exp(x_log_var))
    return kl_loss


# compile and fit
cvae.compile(optimizer=params['optimizer'],
             loss=vae_loss,
             metrics = [KL_loss, recon_loss])

early_stop = EarlyStopping(monitor='val_loss', patience=5, mode='auto')

cvae_hist = cvae.fit([X_train, y_train],
                     X_train,
                     batch_size=512,
                     epochs=18,
                     shuffle='batch',
				        validation_data = ([X_test, y_test], X_test),
                     callbacks=[early_stop])

enc = encoder.predict([X_test,y_test])

model_json = cvae.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvae_publ.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
cvae.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvae_publ.h5")
print("Saved CVAE to disk")

model_json = encoder.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvaeencoder_publ.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
encoder.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvaeencoder_publ.h5")
print("Saved ENCODER to disk")

model_json = decoder.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvaedecoder_publ.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
decoder.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvaedecoder_publ.h5")
print("Saved DECODER to disk")

#cvae.load_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\cvae_publ.h5")

enc = encoder.predict([X_test,y_test])
target_spectra = y_test
err_arg = 1337 #a random number
#generate spectra
x = X_test[err_arg]
z_pred = pred[err_arg]
s = y_test[err_arg]
new_enc = np.copy(enc[err_arg])
encs = []
for i,a in enumerate(a_s):
    for j,k in enumerate(k_s):
        e = np.copy(new_enc)
        e[100:] = tpl(xaxis,0.3,a,k)
        encs.append(e)
encs = np.array(encs)

#predict images from a latent space coordinate with different conditional spectra
g_t = decoder.predict(encs)

fig, axs = plt.subplots(11, 11,figsize=(10,10))
cnt = 0
for i,a in enumerate(a_s):
    for j,k in enumerate(k_s):
        axs[i,j].imshow(g_t[cnt])
        #axs[i,j].plot(tpl(xaxis,0.43,a,k),color='black')
        axs[i,j].axis('off')
        cnt += 1
plt.subplots_adjust(wspace=0, hspace=0)

